@@ -23,72 +23,57 @@ import java.io.ByteArrayOutputStream;
 import java.io.File;
 import java.io.IOException;
 import java.io.InputStream;
+import java.net.MalformedURLException;
+import java.net.URL;
 import java.util.Enumeration;
+import java.util.HashMap;
+import java.util.Map;
+import java.util.Properties;
 import java.util.zip.ZipEntry;
 import org.apache.felix.framework.Logger;
+import org.apache.felix.framework.util.FelixConstants;
 import org.apache.felix.framework.util.JarFileX;
-import org.apache.felix.moduleloader.IContent;
+import org.apache.felix.framework.util.Util;
+import org.apache.felix.framework.resolver.Content;
+import org.osgi.framework.Constants;
 
-public class JarContent implements IContent
+public class JarContent implements Content
 {
     private static final int BUFSIZE = 4096;
-    private static final transient String LEGACY_EMBEDDED_DIRECTORY = "embedded";
     private static final transient String EMBEDDED_DIRECTORY = "-embedded";
-    private static final transient String LIBRARY_DIRECTORY = "lib";
+    private static final transient String LIBRARY_DIRECTORY = "-lib";
 
     private final Logger m_logger;
+    private final Map m_configMap;
     private final Object m_revisionLock;
     private final File m_rootDir;
     private final File m_file;
-    private JarFileX m_jarFile = null;
-    // TODO: CACHE - It would be nice to eventually remove this legacy flag.
-    private final boolean m_legacy;
-    private boolean m_opened = false;
+    private final JarFileX m_jarFile;
+    private final boolean m_isJarFileOwner;
+    private Map m_nativeLibMap;
 
-    public JarContent(Logger logger, Object revisionLock, File rootDir, File file)
+    public JarContent(Logger logger, Map configMap, Object revisionLock, File rootDir,
+        File file, JarFileX jarFile)
     {
         m_logger = logger;
+        m_configMap = configMap;
         m_revisionLock = revisionLock;
         m_rootDir = rootDir;
         m_file = file;
-        m_legacy = false;
-    }
-
-    // This is only used by JarRevision.
-    public JarContent(Logger logger, Object revisionLock, File rootDir, File file, boolean legacy)
-    {
-        m_logger = logger;
-        m_revisionLock = revisionLock;
-        m_rootDir = rootDir;
-        m_file = file;
-        m_legacy = legacy;
+        m_jarFile = (jarFile == null) ? openJarFile(m_file) : jarFile;
+        m_isJarFileOwner = (jarFile == null);
     }
 
     protected void finalize()
     {
-        if (m_jarFile != null)
-        {
-            try
-            {
-                m_jarFile.close();
-            }
-            catch (IOException ex)
-            {
-                // Not much we can do, so ignore it.
-            }
-        }
+        close();
     }
 
-    public synchronized void open()
-    {
-        m_opened = true;
-    }
-
-    public synchronized void close()
+    public void close()
     {
         try
         {
-            if (m_jarFile != null)
+            if (m_isJarFileOwner)
             {
                 m_jarFile.close();
             }
@@ -97,36 +82,12 @@ public class JarContent implements IContent
         {
             m_logger.log(
                 Logger.LOG_ERROR,
-                "JarContent: Unable to open JAR file.", ex);
+                "JarContent: Unable to close JAR file.", ex);
         }
-
-        m_jarFile = null;
-        m_opened = false;
     }
 
-    public synchronized boolean hasEntry(String name) throws IllegalStateException
+    public boolean hasEntry(String name) throws IllegalStateException
     {
-        if (!m_opened)
-        {
-            throw new IllegalStateException("JarContent is not open");
-        }
-
-        // Open JAR file if not already opened.
-        if (m_jarFile == null)
-        {
-            try
-            {
-                openJarFile();
-            }
-            catch (IOException ex)
-            {
-                m_logger.log(
-                    Logger.LOG_ERROR,
-                    "JarContent: Unable to open JAR file.", ex);
-                return false;
-            }
-        }
-
         try
         {
             ZipEntry ze = m_jarFile.getEntry(name);
@@ -141,29 +102,8 @@ public class JarContent implements IContent
         }
     }
 
-    public synchronized Enumeration getEntries()
+    public Enumeration getEntries()
     {
-        if (!m_opened)
-        {
-            throw new IllegalStateException("JarContent is not open");
-        }
-
-        // Open JAR file if not already opened.
-        if (m_jarFile == null)
-        {
-            try
-            {
-                openJarFile();
-            }
-            catch (IOException ex)
-            {
-                m_logger.log(
-                    Logger.LOG_ERROR,
-                    "JarContent: Unable to open JAR file.", ex);
-                return null;
-            }
-        }
-
         // Wrap entries enumeration to filter non-matching entries.
         Enumeration e = new EntriesEnumeration(m_jarFile.entries());
 
@@ -171,29 +111,8 @@ public class JarContent implements IContent
         return (e.hasMoreElements()) ? e : null;
     }
 
-    public synchronized byte[] getEntryAsBytes(String name) throws IllegalStateException
+    public byte[] getEntryAsBytes(String name) throws IllegalStateException
     {
-        if (!m_opened)
-        {
-            throw new IllegalStateException("JarContent is not open");
-        }
-
-        // Open JAR file if not already opened.
-        if (m_jarFile == null)
-        {
-            try
-            {
-                openJarFile();
-            }
-            catch (IOException ex)
-            {
-                m_logger.log(
-                    Logger.LOG_ERROR,
-                    "JarContent: Unable to open JAR file.", ex);
-                return null;
-            }
-        }
-
         // Get the embedded resource.
         InputStream is = null;
         ByteArrayOutputStream baos = null;
@@ -246,30 +165,9 @@ public class JarContent implements IContent
         }
     }
 
-    public synchronized InputStream getEntryAsStream(String name)
+    public InputStream getEntryAsStream(String name)
         throws IllegalStateException, IOException
     {
-        if (!m_opened)
-        {
-            throw new IllegalStateException("JarContent is not open");
-        }
-
-        // Open JAR file if not already opened.
-        if (m_jarFile == null)
-        {
-            try
-            {
-                openJarFile();
-            }
-            catch (IOException ex)
-            {
-                m_logger.log(
-                    Logger.LOG_ERROR,
-                    "JarContent: Unable to open JAR file.", ex);
-                return null;
-            }
-        }
-
         // Get the embedded resource.
         InputStream is = null;
 
@@ -294,28 +192,26 @@ public class JarContent implements IContent
         return is;
     }
 
-    public synchronized IContent getEntryAsContent(String entryName)
+    public URL getEntryAsURL(String name)
     {
-        if (!m_opened)
+        try
         {
-            throw new IllegalStateException("JarContent is not open");
+            return new URL("jar:" + m_file.toURI().toURL().toExternalForm() + "!/" + name);
         }
-
-        // Open JAR file if not already opened.
-        if (m_jarFile == null)
+        catch (MalformedURLException e)
         {
-            try
-            {
-                openJarFile();
-            }
-            catch (IOException ex)
-            {
-                m_logger.log(
-                    Logger.LOG_ERROR,
-                    "JarContent: Unable to open JAR file.", ex);
-                return null;
-            }
+            return null;
+        }
+    }
 
+    public Content getEntryAsContent(String entryName)
+    {
+        // If the entry name refers to the content itself, then
+        // just return it immediately.
+        if (entryName.equals(FelixConstants.CLASS_PATH_DOT))
+        {
+            return new JarContent(m_logger, m_configMap, m_revisionLock,
+                m_rootDir, m_file, m_jarFile);
         }
 
         // Remove any leading slash.
@@ -324,17 +220,7 @@ public class JarContent implements IContent
         // Any embedded JAR files will be extracted to the embedded directory.
         // Since embedded JAR file names may clash when extracting from multiple
         // embedded JAR files, the embedded directory is per embedded JAR file.
-        // For backwards compatibility purposes, don't use the file cache name
-        // for the root bundle JAR file.
-        File embedDir;
-        if (m_legacy)
-        {
-            embedDir = new File(m_rootDir, LEGACY_EMBEDDED_DIRECTORY);
-        }
-        else
-        {
-            embedDir = new File(m_rootDir, m_file.getName() + EMBEDDED_DIRECTORY);
-        }
+        File embedDir = new File(m_rootDir, m_file.getName() + EMBEDDED_DIRECTORY);
 
         // Find the entry in the JAR file and create the
         // appropriate content type for it.
@@ -345,7 +231,7 @@ public class JarContent implements IContent
         ZipEntry ze = m_jarFile.getEntry(entryName);
         if ((ze != null) && ze.isDirectory())
         {
-            File extractedDir = new File(embedDir, entryName);
+            File extractDir = new File(embedDir, entryName);
 
             // Extracting an embedded directory file impacts all other existing
             // contents for this revision, so we have to grab the revision
@@ -353,9 +239,9 @@ public class JarContent implements IContent
             // directory to avoid a race condition.
             synchronized (m_revisionLock)
             {
-                if (!BundleCache.getSecureAction().fileExists(extractedDir))
+                if (!BundleCache.getSecureAction().fileExists(extractDir))
                 {
-                    if (!BundleCache.getSecureAction().mkdirs(extractedDir))
+                    if (!BundleCache.getSecureAction().mkdirs(extractDir))
                     {
                         m_logger.log(
                             Logger.LOG_ERROR,
@@ -367,7 +253,7 @@ public class JarContent implements IContent
         }
         else if ((ze != null) && ze.getName().endsWith(".jar"))
         {
-            File extractedJar = new File(embedDir, entryName);
+            File extractJar = new File(embedDir, entryName);
 
             // Extracting the embedded JAR file impacts all other existing
             // contents for this revision, so we have to grab the revision
@@ -375,112 +261,148 @@ public class JarContent implements IContent
             // to avoid a race condition.
             synchronized (m_revisionLock)
             {
-                if (!BundleCache.getSecureAction().fileExists(extractedJar))
+                try
                 {
-                    try
-                    {
-                        extractEmbeddedJar(entryName);
-                    }
-                    catch (Exception ex)
-                    {
-                        m_logger.log(
-                            Logger.LOG_ERROR,
-                            "Unable to extract embedded JAR file.", ex);
-                    }
+                    extractEmbeddedJar(entryName);
+                }
+                catch (Exception ex)
+                {
+                    m_logger.log(
+                        Logger.LOG_ERROR,
+                        "Unable to extract embedded JAR file.", ex);
                 }
             }
             return new JarContent(
-                m_logger, m_revisionLock, extractedJar.getParentFile(), extractedJar);
+                m_logger, m_configMap, m_revisionLock,
+                extractJar.getParentFile(), extractJar, null);
         }
 
         // The entry could not be found, so return null.
         return null;
     }
 
-// TODO: This will need to consider security.
-    public synchronized String getEntryAsNativeLibrary(String name)
+// TODO: SECURITY - This will need to consider security.
+    public String getEntryAsNativeLibrary(String entryName)
     {
-        if (!m_opened)
-        {
-            throw new IllegalStateException("JarContent is not open");
-        }
+        // Return result.
+        String result = null;
 
-        // Open JAR file if not already opened.
-        if (m_jarFile == null)
-        {
-            try
-            {
-                openJarFile();
-            }
-            catch (IOException ex)
-            {
-                m_logger.log(
-                    Logger.LOG_ERROR,
-                    "JarContent: Unable to open JAR file.", ex);
-                return null;
-            }
-        }
+        // Remove any leading slash.
+        entryName = (entryName.startsWith("/")) ? entryName.substring(1) : entryName;
 
-        // Get bundle lib directory.
-        File libDir = new File(m_rootDir, LIBRARY_DIRECTORY);
-        // Get lib file.
-        File libFile = new File(libDir, File.separatorChar + name);
-        // Make sure that the library's parent directory exists;
-        // it may be in a sub-directory.
-        libDir = libFile.getParentFile();
-        if (!BundleCache.getSecureAction().fileExists(libDir))
-        {
-            if (!BundleCache.getSecureAction().mkdirs(libDir))
-            {
-                m_logger.log(
-                    Logger.LOG_ERROR,
-                    "JarContent: Unable to create library directory.");
-                return null;
-            }
-        }
-        // Extract the library from the JAR file if it does not
-        // already exist.
-        if (!BundleCache.getSecureAction().fileExists(libFile))
-        {
-            InputStream is = null;
+        // Any embedded native libraries will be extracted to the lib directory.
+        // Since embedded library file names may clash when extracting from multiple
+        // embedded JAR files, the embedded lib directory is per embedded JAR file.
+        File libDir = new File(m_rootDir, m_file.getName() + LIBRARY_DIRECTORY);
 
-            try
+        // The entry name must refer to a file type, since it is
+        // a native library, not a directory.
+        ZipEntry ze = m_jarFile.getEntry(entryName);
+        if ((ze != null) && !ze.isDirectory())
+        {
+            // Extracting the embedded native library file impacts all other
+            // existing contents for this revision, so we have to grab the
+            // revision lock first before trying to extract the embedded JAR
+            // file to avoid a race condition.
+            synchronized (m_revisionLock)
             {
-                ZipEntry ze = m_jarFile.getEntry(name);
-                if (ze == null)
+                // Since native libraries cannot be shared, we must extract a
+                // separate copy per request, so use the request library counter
+                // as part of the extracted path.
+                if (m_nativeLibMap == null)
                 {
-                    return null;
-                }
-                is = new BufferedInputStream(
-                    m_jarFile.getInputStream(ze), BundleCache.BUFSIZE);
-                if (is == null)
-                {
-                    throw new IOException("No input stream: " + name);
+                    m_nativeLibMap = new HashMap();
                 }
+                Integer libCount = (Integer) m_nativeLibMap.get(entryName);
+                // Either set or increment the library count.
+                libCount = (libCount == null) ? new Integer(0) : new Integer(libCount.intValue() + 1);
+                m_nativeLibMap.put(entryName, libCount);
+                File libFile = new File(
+                    libDir, libCount.toString() + File.separatorChar + entryName);
 
-                // Create the file.
-                BundleCache.copyStreamToFile(is, libFile);
-            }
-            catch (Exception ex)
-            {
-                m_logger.log(
-                    Logger.LOG_ERROR,
-                    "JarContent: Extracting native library.", ex);
-            }
-            finally
-            {
-                try
+                if (!BundleCache.getSecureAction().fileExists(libFile))
                 {
-                    if (is != null) is.close();
+                    if (!BundleCache.getSecureAction().fileExists(libFile.getParentFile())
+                        && !BundleCache.getSecureAction().mkdirs(libFile.getParentFile()))
+                    {
+                        m_logger.log(
+                            Logger.LOG_ERROR,
+                            "Unable to create library directory.");
+                    }
+                    else
+                    {
+                        InputStream is = null;
+
+                        try
+                        {
+                            is = new BufferedInputStream(
+                                m_jarFile.getInputStream(ze),
+                                BundleCache.BUFSIZE);
+                            if (is == null)
+                            {
+                                throw new IOException("No input stream: " + entryName);
+                            }
+
+                            // Create the file.
+                            BundleCache.copyStreamToFile(is, libFile);
+
+                            // Perform exec permission command on extracted library
+                            // if one is configured.
+                            String command = (String) m_configMap.get(
+                                Constants.FRAMEWORK_EXECPERMISSION);
+                            if (command != null)
+                            {
+                                Properties props = new Properties();
+                                props.setProperty("abspath", libFile.toString());
+                                command = Util.substVars(command, "command", null, props);
+                                Process p = BundleCache.getSecureAction().exec(command);
+                                // We have to make sure we read stdout and stderr because
+                                // otherwise we will block on certain unbuffered os's
+                                // (like eg. windows)
+                                Thread stdOut = new Thread(
+                                    new DevNullRunnable(p.getInputStream()));
+                                Thread stdErr = new Thread(
+                                    new DevNullRunnable(p.getErrorStream()));
+                                stdOut.setDaemon(true);
+                                stdErr.setDaemon(true);
+                                stdOut.start();
+                                stdErr.start();
+                                p.waitFor();
+                                stdOut.join();
+                                stdErr.join();
+                            }
+
+                            // Return the path to the extracted native library.
+                            result = BundleCache.getSecureAction().getAbsolutePath(libFile);
+                        }
+                        catch (Exception ex)
+                        {
+                            m_logger.log(
+                                Logger.LOG_ERROR,
+                                "Extracting native library.", ex);
+                        }
+                        finally
+                        {
+                            try
+                            {
+                                if (is != null) is.close();
+                            }
+                            catch (IOException ex)
+                            {
+                                // Not much we can do.
+                            }
+                        }
+                    }
                 }
-                catch (IOException ex)
+                else
                 {
-                    // Not much we can do.
+                    // Return the path to the extracted native library.
+                    result = BundleCache.getSecureAction().getAbsolutePath(libFile);
                 }
             }
         }
 
-        return BundleCache.getSecureAction().getAbsolutePath(libFile);
+        return result;
     }
 
     public String toString()
@@ -493,14 +415,6 @@ public class JarContent implements IContent
         return m_file;
     }
 
-    private void openJarFile() throws IOException
-    {
-        if (m_jarFile == null)
-        {
-            m_jarFile = BundleCache.getSecureAction().openJAR(m_file);
-        }
-    }
-
     /**
      * This method extracts an embedded JAR file from the bundle's
      * JAR file.
@@ -517,17 +431,7 @@ public class JarContent implements IContent
         // Any embedded JAR files will be extracted to the embedded directory.
         // Since embedded JAR file names may clash when extracting from multiple
         // embedded JAR files, the embedded directory is per embedded JAR file.
-        // For backwards compatibility purposes, don't use the file cache name
-        // for the root bundle JAR file.
-        File embedDir;
-        if (m_legacy)
-        {
-            embedDir = new File(m_rootDir, LEGACY_EMBEDDED_DIRECTORY);
-        }
-        else
-        {
-            embedDir = new File(m_rootDir, m_file.getName() + EMBEDDED_DIRECTORY);
-        }
+        File embedDir = new File(m_rootDir, m_file.getName() + EMBEDDED_DIRECTORY);
         File jarFile = new File(embedDir, jarPath);
 
         if (!BundleCache.getSecureAction().fileExists(jarFile))
@@ -574,9 +478,22 @@ public class JarContent implements IContent
         }
     }
 
+    private static JarFileX openJarFile(File file) throws RuntimeException
+    {
+        try
+        {
+            return BundleCache.getSecureAction().openJAR(file, false);
+        }
+        catch (IOException ex)
+        {
+            throw new RuntimeException(
+                "Unable to open JAR file, probably deleted: " + ex.getMessage());
+        }
+    }
+
     private static class EntriesEnumeration implements Enumeration
     {
-        private Enumeration m_enumeration = null;
+        private final Enumeration m_enumeration;
 
         public EntriesEnumeration(Enumeration enumeration)
         {
@@ -593,4 +510,33 @@ public class JarContent implements IContent
             return ((ZipEntry) m_enumeration.nextElement()).getName();
         }
     }
+
+    private static class DevNullRunnable implements Runnable 
+    {
+        private final InputStream m_in;
+
+        public DevNullRunnable(InputStream in) 
+        {
+            m_in = in;
+        }
+
+        public void run()
+        {
+            try
+            {
+                try
+                {
+                    while (m_in.read() != -1){}
+                }
+                finally 
+                {
+                    m_in.close();
+                }
+            }
+            catch (Exception ex) 
+            {
+                // Not much we can do - maybe we should log it?
+            }
+        }
+    }
 }
\ No newline at end of file
